/* -*- mode: c++; indent-tabs-mode: nil -*- */
/** @file XmlDoc.qpp defines the XmlDoc class */
/*
  QC_XmlDoc.qpp

  Qore Programming Language
  
  Copyright 2003 - 2012 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "qore-xml-module.h"

#include "QC_XmlDoc.h"
#include "QoreXPath.h"
#include "QoreXmlReader.h"
#include "QC_XmlNode.h"
#include "ql_xml.h"

#ifdef HAVE_XMLTEXTREADERRELAXNGSETSCHEMA
int QoreXmlDoc::validateRelaxNG(const char *rng, int size, ExceptionSink *xsink) {
   QoreXmlRelaxNGContext schema(rng, size, xsink);
   if (!schema) {
      if (!*xsink)
	 xsink->raiseException("RELAXNG-SYNTAX-ERROR", "RelaxNG schema passed as argument to XmlDoc::validateRelaxNG() could not be parsed");
      return -1;
   }

   QoreXmlRelaxNGValidContext vcp(schema);
   int rc = vcp.validateDoc(ptr);

   if (!rc)
      return 0;
   if (*xsink)
      return -1;

   if (rc < 0)
      xsink->raiseException("RELAXNG-INTERNAL-ERROR", "an internal error occured validating the document against the RelaxNG schema passed; xmlRelaxNGValidateDoc() returned %d", rc);
   else if (rc)
      xsink->raiseException("RELAXNG-ERROR", "The document failed RelaxNG validation", rc);
   return -1;
}
#endif

#ifdef HAVE_XMLTEXTREADERSETSCHEMA
int QoreXmlDoc::validateSchema(const char *xsd, int size, ExceptionSink *xsink) {
   QoreXmlSchemaContext schema(xsd, size, xsink);
   if (!schema) {
      if (!*xsink)
	 xsink->raiseException("XSD-SYNTAX-ERROR", "XSD schema passed as argument to XmlDoc::validateSchema() could not be parsed");
      return -1;
   }

   QoreXmlSchemaValidContext vcp(schema);
   int rc = vcp.validateDoc(ptr);

   if (!rc)
      return 0;
   if (*xsink)
      return -1;

   if (rc < 0)
      xsink->raiseException("XSD-INTERNAL-ERROR", "an internal error occured validating the document against the XSD schema passed; xmlSchemaValidateDoc() returned %d", rc);
   else if (rc)
      xsink->raiseException("XSD-ERROR", "The document failed XSD validation", rc);
   return -1;
}
#endif

QoreXmlNodeData *QoreXmlDocData::getRootElement() {
   xmlNodePtr n = xmlDocGetRootElement(ptr);
   if (!n) return 0;
   return new QoreXmlNodeData(n, this);
}

QoreStringNode *doString(xmlChar *str) {
   if (!str)
      return 0;
   QoreStringNode *rv = new QoreStringNode((const char *)str);
   xmlFree(str);
   return rv;
}

QoreXmlNodeData *doNode(xmlNodePtr p, QoreXmlDocData *doc) {
   if (!p)
      return 0;
   return new QoreXmlNodeData(p, doc);
}
/* Qore class Qore::Xml::XmlDoc */

qore_classid_t CID_XMLDOC;
QoreClass* QC_XMLDOC;

// XmlDoc::constructor(hash data) {}
static void XmlDoc_constructor_Vh(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreHashNode* data = HARD_QORE_HASH(args, 0);
# 117 "QC_XmlDoc.qpp"
SimpleRefHolder<QoreStringNode> xml(makeXMLString(QCS_UTF8, *data, false, xsink));
   if (!xml)
      return;
   SimpleRefHolder<QoreXmlDocData> xd(new QoreXmlDocData(*xml));
   if (!xd->isValid()) {
      xsink->raiseException("XMLDOC-CONSTRUCTOR-ERROR", "error parsing XML string");
      return;
   }

   self->setPrivate(CID_XMLDOC, xd.release());
}

// XmlDoc::constructor(string xml) {}
static void XmlDoc_constructor_Vs(QoreObject* self, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* xml = HARD_QORE_STRING(args, 0);
# 136 "QC_XmlDoc.qpp"
SimpleRefHolder<QoreXmlDocData> xd(new QoreXmlDocData(xml));
   if (!xd->isValid()) {
      xsink->raiseException("XMLDOC-CONSTRUCTOR-ERROR", "error parsing XML string");
      return;
   }

   self->setPrivate(CID_XMLDOC, xd.release());
}

// XmlDoc::copy() {}
static void XmlDoc_copy(QoreObject* self, QoreObject* old, QoreXmlDocData* xd, ExceptionSink* xsink) {
# 152 "QC_XmlDoc.qpp"
self->setPrivate(CID_XMLDOC, new QoreXmlDocData(*xd));
}

// list XmlDoc::evalXPath(string xpath){}
static AbstractQoreNode* XmlDoc_evalXPath_Vs(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* xpath = HARD_QORE_STRING(args, 0);
# 234 "QC_XmlDoc.qpp"
    QoreXPath xp(xd, xsink);
   if (!xp)
      return 0;

   return xp.eval(xpath->getBuffer(), xsink);
}

// *XmlNode XmlDoc::getRootElement(){}
static QoreObject* XmlDoc_getRootElement(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
# 248 "QC_XmlDoc.qpp"
    QoreXmlNodeData *n = xd->getRootElement();
   return !n ? 0 : new QoreObject(QC_XMLNODE, getProgram(), n);
}

// string XmlDoc::getVersion(){}
static AbstractQoreNode* XmlDoc_getVersion(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
# 162 "QC_XmlDoc.qpp"
    return new QoreStringNode(xd->getVersion());
}

// hash XmlDoc::toQore(){}
static AbstractQoreNode* XmlDoc_toQore(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
# 180 "QC_XmlDoc.qpp"
    QoreXmlReader reader(xd->getDocPtr(), xsink);
   if (*xsink)
      return 0;
   return reader.parseXMLData(QCS_UTF8, false, xsink);
}

// hash XmlDoc::toQoreData(){}
static AbstractQoreNode* XmlDoc_toQoreData(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
# 203 "QC_XmlDoc.qpp"
    QoreXmlReader reader(xd->getDocPtr(), xsink);
   if (*xsink)
      return 0;
   return reader.parseXMLData(QCS_UTF8, true, xsink);
}

// string XmlDoc::toString(){}
static AbstractQoreNode* XmlDoc_toString(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
# 219 "QC_XmlDoc.qpp"
    return xd->toString(xsink);
}

// nothing XmlDoc::validateRelaxNG(string relaxng){}
static AbstractQoreNode* XmlDoc_validateRelaxNG_Vs(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* relaxng = HARD_QORE_STRING(args, 0);
# 265 "QC_XmlDoc.qpp"
    #ifdef HAVE_XMLTEXTREADERRELAXNGSETSCHEMA
   // convert to UTF-8 
   TempEncodingHelper nrng(relaxng, QCS_UTF8, xsink);
   if (!nrng)
      return 0;

   xd->validateRelaxNG(nrng->getBuffer(), nrng->strlen(), xsink);
#else
   xsink->raiseException("MISSING-FEATURE-ERROR", "the libxml2 version used to compile the xml module did not support the xmlTextReaderRelaxNGValidate() function, therefore XmlDoc::validateRelaxNG() is not available; for maximum portability, use the constant Option::HAVE_PARSEXMLWITHRELAXNG to check if this method is implemented before calling");
#endif
   return 0;
}

// nothing XmlDoc::validateSchema(string xsd){}
static AbstractQoreNode* XmlDoc_validateSchema_Vs(QoreObject* self, QoreXmlDocData* xd, const QoreListNode* args, ExceptionSink* xsink) {
    const QoreStringNode* xsd = HARD_QORE_STRING(args, 0);
# 292 "QC_XmlDoc.qpp"
    #ifdef HAVE_XMLTEXTREADERSETSCHEMA
   // convert to UTF-8 
   TempEncodingHelper nxsd(xsd, QCS_UTF8, xsink);
   if (!nxsd)
      return 0;

   xd->validateSchema(nxsd->getBuffer(), nxsd->strlen(), xsink);
#else
   xsink->raiseException("MISSING-FEATURE-ERROR", "the libxml2 version used to compile the xml module did not support the xmlTextReaderSchemaValidate() function, therefore XmlDoc::validateSchema() is not available; for maximum portability, use the constant Option::HAVE_PARSEXMLWITHSCHEMA to check if this method is implemented before calling");
#endif
   return 0;
}

DLLLOCAL void preinitXmlDocClass() {
    QC_XMLDOC = new QoreClass("XmlDoc", QDOM_DEFAULT);
    CID_XMLDOC = QC_XMLDOC->getID();
    QC_XMLDOC->setSystem();
}

DLLLOCAL QoreClass* initXmlDocClass(QoreNamespace& ns) {
    if (!QC_XMLDOC)
        preinitXmlDocClass();

    // XmlDoc::constructor(hash data) {}
    QC_XMLDOC->setConstructorExtended3(XmlDoc_constructor_Vh, false, QC_NO_FLAGS, QDOM_DEFAULT, 1, hashTypeInfo, NULL, "data");

    // XmlDoc::constructor(string xml) {}
    QC_XMLDOC->setConstructorExtended3(XmlDoc_constructor_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, 1, stringTypeInfo, NULL, "xml");

    // XmlDoc::copy() {}
    QC_XMLDOC->setCopy((q_copy_t)XmlDoc_copy);

    // list XmlDoc::evalXPath(string xpath){}
    QC_XMLDOC->addMethodExtended3("evalXPath", (q_method_t)XmlDoc_evalXPath_Vs, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, listTypeInfo, 1, stringTypeInfo, NULL, "xpath");

    // *XmlNode XmlDoc::getRootElement(){}
    QC_XMLDOC->addMethodExtended3("getRootElement", (q_method_t)XmlDoc_getRootElement, false, QC_CONSTANT, QDOM_DEFAULT, QC_XMLNODE->getOrNothingTypeInfo());

    // string XmlDoc::getVersion(){}
    QC_XMLDOC->addMethodExtended3("getVersion", (q_method_t)XmlDoc_getVersion, false, QC_CONSTANT, QDOM_DEFAULT, stringTypeInfo);

    // hash XmlDoc::toQore(){}
    QC_XMLDOC->addMethodExtended3("toQore", (q_method_t)XmlDoc_toQore, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashTypeInfo);

    // hash XmlDoc::toQoreData(){}
    QC_XMLDOC->addMethodExtended3("toQoreData", (q_method_t)XmlDoc_toQoreData, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, hashTypeInfo);

    // string XmlDoc::toString(){}
    QC_XMLDOC->addMethodExtended3("toString", (q_method_t)XmlDoc_toString, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

    // nothing XmlDoc::validateRelaxNG(string relaxng){}
    QC_XMLDOC->addMethodExtended3("validateRelaxNG", (q_method_t)XmlDoc_validateRelaxNG_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "relaxng");

    // nothing XmlDoc::validateSchema(string xsd){}
    QC_XMLDOC->addMethodExtended3("validateSchema", (q_method_t)XmlDoc_validateSchema_Vs, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, NULL, "xsd");

    return QC_XMLDOC;
}
